//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

//==========================================================================
//
//	A_Look
//
//	Stay in state until a player is sighted
//
//==========================================================================

void A_Look(mobj_t* actor)
{
	mobj_t*		targ;

	actor->threshold = 0;       // any shot will wake up
	targ = actor->subsector->sector->soundtarget;

	if (targ && targ->flags & MF_SHOOTABLE)
	{
		actor->target = targ;

		if ( actor->flags & MF_AMBUSH )
		{
			if (!P_CheckSight(actor, actor->target))
			{
				if (!P_LookForPlayers (actor, false) )
					return;
			}
		}
	}
	else
	{
		if (!P_LookForPlayers(actor, false))
			return;
	}

	// go into chase state
	if (actor->sound_sight)
	{
		if (actor->flags2 & MF2_BOSS)
		{ // Full volume
			StartSound(NULL, actor->sound_sight);
		}
		else
		{
			StartSound(actor, actor->sound_sight);
		}
	}

	SetMobjState(actor, actor->seestate);
}

//==========================================================================
//
//	A_Chase
//
//	Actor has a melee attack, so it tries to close as fast as possible
//
//==========================================================================

void A_Chase(mobj_t* actor)
{
    int		delta;

	if(actor->reactiontime)
	{
		actor->reactiontime--;
	}

    // modify target threshold
    if  (actor->threshold)
    {
		if (!actor->target || actor->target->health <= 0)
		{
		    actor->threshold = 0;
		}
		else
        {
		    actor->threshold--;
		}
	}

	if (fastparm)
	{
		// Monsters move faster in nightmare mode
		actor->time *= 0.5;
		if (actor->time < 0.1)
		{
			actor->time = 0.1;
		}
	}

    // turn towards movement direction if not there yet
    if (actor->movedir < 8)
    {
		actor->angles.yaw &= (7 << 29);
		delta = actor->angles.yaw - (actor->movedir << 29);
	
		if (delta > 0)
        {
		    actor->angles.yaw -= ANG90 / 2;
		}
		else if (delta < 0)
		{
		    actor->angles.yaw += ANG90 / 2;
		}
    }

    if (!actor->target || !(actor->target->flags & MF_SHOOTABLE))
    {
		// look for a new target
		if (P_LookForPlayers(actor,true))
        {
		    return; 	// got a new target
	    }
		SetMobjState(actor, actor->spawnstate);
		return;
    }
    
    // do not attack twice in a row
    if (actor->flags & MF_JUSTATTACKED)
    {
		actor->flags &= ~MF_JUSTATTACKED;
		if (!fastparm)
        {
		    P_NewChaseDir(actor);
		}
		return;
    }
    
    // check for melee attack
    if (actor->meleestate	&& P_CheckMeleeRange(actor))
    {
		if (actor->sound_attack)
        {
		    StartSound(actor, actor->sound_attack);
		}
		SetMobjState(actor, actor->meleestate);
		return;
    }
    
    // check for missile attack
    if (actor->missilestate)
    {
		if (fastparm || !actor->movecount)
        {
			if (P_CheckMissileRange(actor))
            {
				SetMobjState(actor, actor->missilestate);
				actor->flags |= MF_JUSTATTACKED;
				return;
			}
		}
    }

    // possibly choose another target
    if (netgame && !actor->threshold)
    {
    	if (!P_CheckSight(actor, actor->target))
	    {
			if (P_LookForPlayers(actor,true))
			    return;	// got a new target
	    }
	}

    // chase towards player
    if (--actor->movecount < 0 || !P_Move(actor))
    {
		P_NewChaseDir(actor);
    }

    // make active sound
    if (actor->sound_active && P_Random() < 3)
    {
		if (actor->type == MT_WIZARD && P_Random() < 128)
		{
			StartSound(actor, actor->sound_sight);
		}
		else if (actor->type == MT_SORCERER2)
		{
			StartSound(NULL, actor->sound_active);
		}
		else
		{
			StartSound(actor, actor->sound_active);
		}
    }
}

//==========================================================================
//
//	A_FaceTarget
//
//==========================================================================

void A_FaceTarget(mobj_t* actor)
{
    if (!actor->target)
		return;
    
    actor->flags &= ~MF_AMBUSH;
	
    actor->angles.yaw = atan2(actor->target->origin.y - actor->origin.y,
	    actor->target->origin.x - actor->origin.x);
    
    if (actor->target->translucency)
	{
		actor->angles.yaw += (P_Random() - P_Random()) << 21;
	}
}

//==========================================================================
//
//	A_Pain
//
//==========================================================================

void A_Pain(mobj_t* actor)
{
    if (actor->sound_pain)
    {
		StartSound(actor, actor->sound_pain);
	}
}

//==========================================================================
//
//	A_FreeTargMobj
//
//==========================================================================

void A_FreeTargMobj(mobj_t* mo)
{
	mo->velocity.x = mo->velocity.y = mo->velocity.z = 0.0;
	mo->origin.z = mo->ceilingz + 4.0;
	mo->flags &= ~(MF_SHOOTABLE | MF_FLOAT | MF_SKULLFLY | MF_SOLID);
	mo->flags |= MF_CORPSE | MF_DROPOFF | MF_NOGRAVITY;
	mo->flags2 &= ~(MF2_PASSMOBJ | MF2_LOGRAV);
	mo->player = NULL;
}

//==========================================================================
//
//	A_Scream
//
//==========================================================================

void A_Scream(mobj_t* actor)
{
	switch(actor->type)
	{
		case MT_CHICPLAYER:
		case MT_SORCERER1:
		case MT_MINOTAUR:
			// Make boss death sounds full volume
			StartSound(NULL, actor->sound_death);
			break;
		case MT_PLAYER:
			// Handle the different player death screams
			if(actor->special1 < 10)
			{ // Wimpy death sound
				StartSoundName(actor, "PlayerWimpyDeath");
			}
			else if(actor->health > -50)
			{ // Normal death sound
				StartSound(actor, actor->sound_death);
			}
			else if(actor->health > -100)
			{ // Crazy death sound
				StartSoundName(actor, "PlayerCrazyDeath");
			}
			else
			{ // Extreme death sound
				StartSoundName(actor, "PlayerExtremeDeath");
			}
			break;
		default:
			StartSound(actor, actor->sound_death);
			break;
	}
}

//==========================================================================
//
//	A_Explode
//
//	Handles a bunch of exploding things.
//
//==========================================================================

void A_Explode(mobj_t* actor)
{
	int 	damage;

	damage = 128;
	switch (actor->type)
	{
		case MT_FIREBOMB: // Time Bombs
			actor->origin.z += 32.0;
			actor->translucency = 0;
			break;
		case MT_MNTRFX2: // Minotaur floor fire
			damage = 24;
			break;
		case MT_SOR2FX1: // D'Sparil missile
			damage = 80 + (P_Random() & 31);
			break;
		default:
			break;
	}
	P_RadiusAttack(actor, actor->target, damage, 0, 0);
	P_HitFloor(actor);
}

//==========================================================================
//
//	P_DropItem
//
//==========================================================================

void P_DropItem(mobj_t* source, int type, int special, int chance)
{
	mobj_t*	mo;

	if (P_Random() > chance)
	{
		return;
	}
	mo = P_SpawnMobj(source->origin.x, source->origin.y,
		source->origin.z + source->height / 2.0, type);
	mo->velocity.x = (Random() - Random()) * 35.0;
	mo->velocity.y = (Random() - Random()) * 35.0;
	mo->velocity.z = (5.0 + Random() * 4.0) * 35.0;
	mo->flags |= MF_DROPPED;
	mo->health = special;
}

//==========================================================================
//
//	A_NoBlocking
//
//==========================================================================

void A_NoBlocking(mobj_t* actor)
{
	actor->flags &= ~MF_SOLID;
	// Check for monsters dropping things
	switch(actor->type)
	{
		case MT_MUMMY:
		case MT_MUMMYLEADER:
		case MT_MUMMYGHOST:
		case MT_MUMMYLEADERGHOST:
			P_DropItem(actor, MT_AMGWNDWIMPY, 3, 84);
			break;
		case MT_KNIGHT:
		case MT_KNIGHTGHOST:
			P_DropItem(actor, MT_AMCBOWWIMPY, 5, 84);
			break;
		case MT_WIZARD:
			P_DropItem(actor, MT_AMBLSRWIMPY, 10, 84);
			P_DropItem(actor, MT_ARTITOMEOFPOWER, 0, 4);
			break;
		case MT_HEAD:
			P_DropItem(actor, MT_AMBLSRWIMPY, 10, 84);
			P_DropItem(actor, MT_ARTIEGG, 0, 51);
			break;
		case MT_BEAST:
			P_DropItem(actor, MT_AMCBOWWIMPY, 10, 84);
			break;
		case MT_CLINK:
			P_DropItem(actor, MT_AMSKRDWIMPY, 20, 84);
			break;
		case MT_SNAKE:
			P_DropItem(actor, MT_AMPHRDWIMPY, 5, 84);
			break;
		case MT_MINOTAUR:
			P_DropItem(actor, MT_ARTISUPERHEAL, 0, 51);
			P_DropItem(actor, MT_AMPHRDWIMPY, 10, 84);
			break;
		default:
			break;
	}
}

//==========================================================================
//
// PROC A_AddPlayerCorpse
//
//==========================================================================

#define BODYQUESIZE		32
mobj_t*	bodyque[BODYQUESIZE];
int 	bodyqueslot;

void A_AddPlayerCorpse(mobj_t* actor)
{
	if (bodyqueslot >= BODYQUESIZE)
	{
		// Too many player corpses - remove an old one
		P_RemoveMobj(bodyque[bodyqueslot % BODYQUESIZE]);
	}
	bodyque[bodyqueslot % BODYQUESIZE] = actor;
	bodyqueslot++;
}

//==========================================================================
//
//	A_SkullPop
//
//==========================================================================

void A_SkullPop(mobj_t* actor)
{
	mobj_t*		mo;
	player_t*	player;

	actor->flags &= ~MF_SOLID;
	mo = P_SpawnMobj(actor->origin.x, actor->origin.y, actor->origin.z + 48.0, MT_BLOODYSKULL);
	//mo->target = actor;
	mo->velocity.x = (Random() - Random()) * 2.0 * 35.0;
	mo->velocity.y = (Random() - Random()) * 2.0 * 35.0;
	mo->velocity.z = (2.0 + Random() / 4.0) * 35.0;
	// Attach player mobj to bloody skull
	player = actor->player;
	actor->player = NULL;
	mo->player = player;
	mo->health = actor->health;
	mo->angles.yaw = actor->angles.yaw;
	player->mo = mo;
	player->damagecount = 32;
}

//==========================================================================
//
// PROC A_FlameSnd
//
//==========================================================================

void A_FlameSnd(mobj_t* actor)
{
	StartSoundName(actor, "HeadAttack1"); // Burn sound
}

//==========================================================================
//
//	A_CheckBurnGone
//
//==========================================================================

void A_CheckBurnGone(mobj_t* actor)
{
	if(actor->special2 == 666)
	{
		SetMobjState(actor, S_PLAY_FDTH20);
	}
}

//==========================================================================
//
// PROC A_CheckSkullFloor
//
//==========================================================================

void A_CheckSkullFloor(mobj_t* actor)
{
	if (actor->origin.z <= actor->floorz)
	{
		SetMobjState(actor, S_BLOODYSKULLX1);
	}
}

//==========================================================================
//
// PROC A_CheckSkullDone
//
//==========================================================================

void A_CheckSkullDone(mobj_t* actor)
{
	if (actor->special2 == 666)
	{
		SetMobjState(actor, S_BLOODYSKULLX2);
	}
}

//==========================================================================
//
//	A_BossDeath
//
//	Trigger special effects if all bosses are dead.
//
//==========================================================================

void A_BossDeath(mobj_t *actor)
{
	string		mapname;
	mobj_t		*mo;
	int			args[5];

	mapname = ptrtos(&level->mapname[0]);

	if (!strcmp(mapname, "E1M8"))
	{
	    if (actor->type != MT_HEAD)
		{
			// Not considered a boss in this episode
			return;
		}
	}
	else if (!strcmp(mapname, "E2M8"))
	{
	    if (actor->type != MT_MINOTAUR)
		{
			// Not considered a boss in this episode
			return;
		}
	}
	else if (!strcmp(mapname, "E3M8"))
	{
	    if (actor->type != MT_SORCERER2)
		{
			// Not considered a boss in this episode
			return;
		}
	}
	else if (!strcmp(mapname, "E4M8"))
	{
	    if (actor->type != MT_HEAD)
		{
			// Not considered a boss in this episode
			return;
		}
	}
	else if (!strcmp(mapname, "E5M8"))
	{
	    if (actor->type != MT_MINOTAUR)
		{
			// Not considered a boss in this episode
			return;
		}
	}
	else
	{
		//	Not a boss level
		return;
	}

	// Make sure all bosses are dead
	for (mo = NextMobj(NULL); mo; mo = NextMobj(mo))
	{
		if ((mo != actor) && (mo->type == actor->type) && (mo->health > 0))
		{
			// Found a living boss
			return;
		}
	}
	if (strcmp(mapname, "E1M8"))
	{
		// Kill any remaining monsters
		P_Massacre();
	}
	args[0] = 666;
	args[1] = 8;
	EV_DoFloor(NULL, &args[0], FLEV_LOWERFLOOR);
}

//==========================================================================
//
//	A_ContMobjSound
//
//==========================================================================

void A_ContMobjSound(mobj_t* actor)
{
	switch (actor->type)
	{
		case MT_KNIGHTAXE:
			StartSoundName(actor, "KnightAttack1");
			break;
		case MT_MUMMYFX1:
			StartSoundName(actor, "MumyHeadExplode");
			break;  
		default:
			break;
	}
}


