//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**
//**	Functions to draw patches (by post) directly to screen.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d3d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TDirect3DDrawer::Setup2D
//
//==========================================================================

void TDirect3DDrawer::Setup2D(void)
{
	//	Setup viewport
	D3DVIEWPORT7 view2D;
	memset(&view2D, 0, sizeof(D3DVIEWPORT7));
	view2D.dwX = 0;
	view2D.dwY = 0;
	view2D.dwWidth  = ScreenWidth;
	view2D.dwHeight = ScreenHeight;
	view2D.dvMinZ = 0.0f;
	view2D.dvMaxZ = 1.0f;
    RenderDevice->SetViewport(&view2D);

	//	Setup projection
	D3DMATRIX proj2D = IdentityMatrix;
	proj2D(0, 0) = 2.0 / (float)ScreenWidth;
	proj2D(1, 1) = -2.0 / (float)ScreenHeight;
	proj2D(3, 0) = -1.0;
	proj2D(3, 1) = 1.0;
	RenderDevice->SetTransform(D3DTRANSFORMSTATE_PROJECTION, &proj2D);

	RenderDevice->SetTransform(D3DTRANSFORMSTATE_VIEW, &IdentityMatrix);

	RenderDevice->SetRenderState(D3DRENDERSTATE_SHADEMODE, D3DSHADE_FLAT);
	RenderDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, FALSE);
	RenderDevice->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE);
	RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, FALSE);
	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
	}
	else
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawPic
//
//==========================================================================

void TDirect3DDrawer::DrawPic(float x1, float x2, float y1, float y2,
	float s1, float s2, float t1, float t2, int handle, int trans)
{
	D3DLVERTEX	dv[4];
	int l = (((100 - trans) * 255 / 100) << 24) | 0xffffff;

	SetPic(handle);

	dv[0] = D3DLVERTEX(D3DVECTOR(x1, y1, 0), l, 0, s1 * tex_iw, t1 * tex_ih);
	dv[1] = D3DLVERTEX(D3DVECTOR(x2, y1, 0), l, 0, s2 * tex_iw, t1 * tex_ih);
	dv[2] = D3DLVERTEX(D3DVECTOR(x2, y2, 0), l, 0, s2 * tex_iw, t2 * tex_ih);
	dv[3] = D3DLVERTEX(D3DVECTOR(x1, y2, 0), l, 0, s1 * tex_iw, t2 * tex_ih);

	if (trans && HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, dv, 4, 0);

	if (trans && HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawPicShadow
//
//==========================================================================

void TDirect3DDrawer::DrawPicShadow(float x1, float x2, float y1, float y2,
	float s1, float s2, float t1, float t2, int handle, int shade)
{
	D3DLVERTEX	dv[4];
	int l = shade << 24;

	SetPic(handle);

	dv[0] = D3DLVERTEX(D3DVECTOR(x1, y1, 0), l, 0, s1 * tex_iw, t1 * tex_ih);
	dv[1] = D3DLVERTEX(D3DVECTOR(x2, y1, 0), l, 0, s2 * tex_iw, t1 * tex_ih);
	dv[2] = D3DLVERTEX(D3DVECTOR(x2, y2, 0), l, 0, s2 * tex_iw, t2 * tex_ih);
	dv[3] = D3DLVERTEX(D3DVECTOR(x1, y2, 0), l, 0, s1 * tex_iw, t2 * tex_ih);

	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, dv, 4, 0);

	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	}
}

//==========================================================================
//
//  TDirect3DDrawer::FillRectWithFlat
//
// 	Fills rectangle with flat.
//
//==========================================================================

void TDirect3DDrawer::FillRectWithFlat(int x, int y, int w, int h, const char* fname)
{
	D3DLVERTEX	dv[4];
	int l = 0xffffffff;

	SetFlat(R_FlatNumForName(fname));

	dv[0] = D3DLVERTEX(D3DVECTOR(x, y, 0), l, 0, 0, 0);
	dv[1] = D3DLVERTEX(D3DVECTOR(x + w, y, 0), l, 0, w / 64.0, 0);
	dv[2] = D3DLVERTEX(D3DVECTOR(x + w, y + h, 0), l, 0, w / 64.0, h / 64.0);
	dv[3] = D3DLVERTEX(D3DVECTOR(x, y + h, 0), l, 0, 0, h / 64.0);

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, dv, 4, 0);
}

//==========================================================================
//
//	TDirect3DDrawer::DarkenScreen
//
//  Fade all the screen buffer, so that the menu is more readable,
// especially now that we use the small hufont in the menus...
//
//==========================================================================

void TDirect3DDrawer::ShadeRect(int x, int y, int w, int h, int darkening)
{
	D3DLVERTEX	dv[4];
	int l = darkening << 27;

	RenderDevice->SetTexture(0, NULL);
	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	}

	dv[0] = D3DLVERTEX(D3DVECTOR(x, y, 0), l, 0, 0, 0);
	dv[1] = D3DLVERTEX(D3DVECTOR(x + w, y, 0), l, 0, 0, 0);
	dv[2] = D3DLVERTEX(D3DVECTOR(x + w, y + h, 0), l, 0, 0, 0);
	dv[3] = D3DLVERTEX(D3DVECTOR(x, y + h, 0), l, 0, 0, 0);

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, dv, 4, 0);

	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawConsoleBackground
//
//==========================================================================

void TDirect3DDrawer::DrawConsoleBackground(int h)
{
	D3DLVERTEX	dv[4];
	int l = 0xc000007f;

	RenderDevice->SetTexture(0, NULL);
	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	}

	dv[0] = D3DLVERTEX(D3DVECTOR(0, 0, 0), l, 0, 0, 0);
	dv[1] = D3DLVERTEX(D3DVECTOR(ScreenWidth, 0, 0), l, 0, 0, 0);
	dv[2] = D3DLVERTEX(D3DVECTOR(ScreenWidth, h, 0), l, 0, 0, 0);
	dv[3] = D3DLVERTEX(D3DVECTOR(0, h, 0), l, 0, 0, 0);

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, dv, 4, 0);

	if (HavesAlphaTest)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
		RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
	}
}

//==========================================================================
//
//	TDirect3DDrawer::DrawSpriteLump
//
//==========================================================================

void TDirect3DDrawer::DrawSpriteLump(int x, int y, int lump, int translation, boolean flip)
{
	SetSpriteLump(lump, translation);

    x -= (int)spriteoffset[lump];
   	y -= (int)spritetopoffset[lump];
	float w = spritewidth[lump];
	float h = spriteheight[lump];
	float texw = w * tex_iw;
	float texh = h * tex_ih;

	D3DLVERTEX	dv[4];
	int l = 0xffffffff;

	if (flip)
	{
		dv[0] = D3DLVERTEX(D3DVECTOR(x, y, 0), l, 0, texw, 0);
		dv[1] = D3DLVERTEX(D3DVECTOR(x + w, y, 0), l, 0, 0, 0);
		dv[2] = D3DLVERTEX(D3DVECTOR(x + w, y + h, 0), l, 0, 0, texh);
		dv[3] = D3DLVERTEX(D3DVECTOR(x, y + h, 0), l, 0, texw, texh);
	}
	else
	{
		dv[0] = D3DLVERTEX(D3DVECTOR(x, y, 0), l, 0, 0, 0);
		dv[1] = D3DLVERTEX(D3DVECTOR(x + w, y, 0), l, 0, texw, 0);
		dv[2] = D3DLVERTEX(D3DVECTOR(x + w, y + h, 0), l, 0, texw, texh);
		dv[3] = D3DLVERTEX(D3DVECTOR(x, y + h, 0), l, 0, 0, texh);
	}

	RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, dv, 4, 0);
}


