//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**
//**	Local header for Direct3D drawer
//**
//**************************************************************************

#ifndef _D3D_LOCAL_H
#define _D3D_LOCAL_H

// HEADER FILES ------------------------------------------------------------

#define D3D_OVERLOADS
#include "winlocal.h"
#include <d3d.h>
#include "gamedefs.h"
#include "cl_local.h"
#include "r_shared.h"

// MACROS ------------------------------------------------------------------

#define SAFE_RELEASE(iface) \
	if (iface) \
	{ \
		iface->Release(); \
		iface = NULL; \
	}

// Lighting constants.
// Now why not 32 levels here?
#define LIGHTLEVELS		16
#define LIGHTSEGSCALE	0.0625

#define MAXLIGHTZ		128

#define MAX_TRANSLATED_SPRITES		256

// TYPES -------------------------------------------------------------------

class TDirect3DDrawer : public TDrawer
{
 public:
	TDirect3DDrawer(void);
	void Init(void);
	void InitData(void);
	bool SetResolution(int, int, int);
	void InitResolution(void);
	void SetPalette(int);
	void StartUpdate(void);
	void Update(void);
	void Shutdown(void);
	void* ReadScreen(int*, bool*);

	//	Screen wipes
	bool InitWipe(void);
	void DoWipe(int);

	//	Rendering stuff
	void SetupView(int, int, int, int, float, float);
	void SetupFrame(void);
	void WorldDrawing(void);
	void EndView(void);

	//	Texture stuff
	void InitTextures(void);
	void SetTexture(int);
	void SetSkyTexture(int, bool);
	void SetFlat(int);
	void SetSpriteLump(int, int);

	//	Polygon drawing
	void DrawPolygon(draw_vertex_t *, int, int, int, bool);
	void DrawMaskedPolygon(draw_vertex_t *, int, int, int, int, bool);
	void DrawZPolygon(draw_vertex_t *, int, int);
	void DrawSkyPolygon(draw_vertex_t *, int, int, bool);
	void DrawSpritePolygon(draw_vertex_t *, int, int, int, bool);

	//	Drawing
	void DrawPic(float, float, float, float, float, float, float, float, int, int);
	void DrawPicShadow(float, float, float, float, float, float, float, float, int, int);
	void FillRectWithFlat(int, int, int, int, const char*);
	void ShadeRect(int, int, int, int, int);
	void DrawConsoleBackground(int);
	void DrawSpriteLump(int, int, int, int, boolean);

 private:
	void InitLightTables(void);
	void EmitVertex(D3DLVERTEX &lv, const draw_vertex_t &dv, byte alpha);

	void Setup2D(void);

	void ReleaseTextures(void);
	int ToPowerOf2(int);
	LPDIRECTDRAWSURFACE7 CreateSurface(int, int);
	word *LockSurface(LPDIRECTDRAWSURFACE7);
	void DrawColumnInCache(column_t*, word*, int, int, int, int);
	void GenerateTexture(int);
	void GenerateFlat(int);
	void GenerateSprite(int);
	void GenerateTranslatedSprite(int, int, int);
	void SetPic(int);
	void GeneratePicFromPatch(int);
	void GeneratePicFromRaw(int);

	word MakeCol16(byte r, byte g, byte b)
	{
		return abit |
			((r >> 3) << rshift) |
			((g >> 3) << gshift) |
			((b >> 3) << bshift);
	}

	//	DirectDraw interfaces
	LPDIRECTDRAW7				DDraw;
	LPDIRECTDRAWSURFACE7		PrimarySurface;
	LPDIRECTDRAWSURFACE7		RenderSurface;
	LPDIRECTDRAWSURFACE7		ZBuffer;

	//	Direct3D interfaces
	LPDIRECT3D7					Direct3D;
	LPDIRECT3DDEVICE7			RenderDevice;

	D3DMATRIX					IdentityMatrix;

	D3DVIEWPORT7				viewData;
	D3DMATRIX					matProj;

	DDPIXELFORMAT				PixelFormat;
	dword						SurfaceMemFlag;
	bool						HavesAlphaTest;

	int							abit;
	int							rshift;
	int							gshift;
	int							bshift;

	word						pal8_to16[256];

	float						tex_iw;
	float						tex_ih;

	int							maxTexSize;

	LPDIRECTDRAWSURFACE7		*texturesurfaces;
	int							numsurfaces;
	int							tscount;

	LPDIRECTDRAWSURFACE7		*texturedata;
	float						*textureiw;
	float						*textureih;
	LPDIRECTDRAWSURFACE7		*flatdata;
	LPDIRECTDRAWSURFACE7		*spritedata;
	float						*spriteiw;
	float						*spriteih;
	LPDIRECTDRAWSURFACE7		*trsprdata;
	int							trsprlump[MAX_TRANSLATED_SPRITES];
	int							trsprtnum[MAX_TRANSLATED_SPRITES];
	float						trspriw[MAX_TRANSLATED_SPRITES];
	float						trsprih[MAX_TRANSLATED_SPRITES];

	LPDIRECTDRAWSURFACE7		*picdata;
	float						piciw[MAX_PICS];
	float						picih[MAX_PICS];

	//	Lighting LUT.
	//	Used for z-depth cuing per column/row,
	// and other lighting effects (sector ambient, flash).
	float						zlight[LIGHTLEVELS][MAXLIGHTZ];

	TCvarI						use_hal;
	TCvarI						clear;
	TCvarI						tex_linear;
	TCvarI						dither;
	TCvarF						light_z_scale;
	TCvarI						use_zlight;
	TCvarI						square_textures;
};

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PUBLIC DATA DECLARATIONS ------------------------------------------------

#endif

