//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d3d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

extern byte					gammatable[5][256];

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

ostream &operator << (ostream &str, const LPD3DDEVICEDESC7 dd);
ostream &operator << (ostream &str, const LPDDPIXELFORMAT pf);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static TDirect3DDrawer		Direct3DDrawer;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TDirect3DDrawer::TDirect3DDrawer
//
//==========================================================================

TDirect3DDrawer::TDirect3DDrawer(void) :
	use_hal("d3d_use_hal", "1", CVAR_ARCHIVE),
	clear("d3d_clear", "0", CVAR_ARCHIVE),
	tex_linear("d3d_tex_linear", "1", CVAR_ARCHIVE),
	dither("d3d_dither", "0", CVAR_ARCHIVE),
	light_z_scale("d3d_light_z_scale", "0.0625"),
	use_zlight("d3d_use_zlight", "1", CVAR_ARCHIVE),
	square_textures("d3d_square_textures", "1", CVAR_ARCHIVE),
	IdentityMatrix(	1, 0, 0, 0,
					0, 1, 0, 0,
					0, 0, 1, 0,
					0, 0, 0, 1)
{
	_Direct3DDrawer = this;

	//	DirectDraw interfaces
	DDraw = NULL;
	PrimarySurface = NULL;
	RenderSurface = NULL;
	ZBuffer = NULL;

	//	Direct3D interfaces
	Direct3D = NULL;
	RenderDevice = NULL;
}

//==========================================================================
//
//	TDirect3DDrawer::Init
//
// 	Determine the hardware configuration
//
//==========================================================================

void TDirect3DDrawer::Init(void)
{
	HRESULT			result;

	// Create DirectDraw object
	result = CoCreateInstance(CLSID_DirectDraw7, NULL,
		CLSCTX_ALL, IID_IDirectDraw7, (void**)&DDraw);
	if (result != DD_OK)
		I_Error("I_InitGraphics: Failed to create DirecDraw7");

	// Initialize
	result = DDraw->Initialize(NULL);
	if (result != DD_OK)
		I_Error("I_InitGraphics: Failed to initialize DirectDraw");

	// Set cooperative level
	result = DDraw->SetCooperativeLevel(hwnd,
		DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
	if (result != DD_OK)
		I_Error("I_InitGraphics: Failed to set cooperative level.");

	// Create Direct3D object
	result = DDraw->QueryInterface(IID_IDirect3D7, (void**)&Direct3D);
	if (FAILED(result))
		I_Error("Failed to create Direct3D object");
}

//==========================================================================
//
//	TDirect3DDrawer::InitData
//
//==========================================================================

void TDirect3DDrawer::InitData(void)
{
    InitLightTables();
}

//==========================================================================
//
//	EnumDevicesCallback
//
//==========================================================================

static HRESULT CALLBACK EnumDevicesCallback(
	LPSTR lpDeviceDesc,
	LPSTR lpDeviceName,
	LPD3DDEVICEDESC7 lpD3DDeviceDesc,
	LPVOID lpContext)
{
	con << "Device " << lpDeviceName << endl;
	con << "Description: " << lpDeviceDesc << endl;
	con << lpD3DDeviceDesc;
	con << "-------------------------------------\n";

	return D3DENUMRET_OK;
}

//==========================================================================
//
//	EnumZBufferCallback
//
//==========================================================================

static HRESULT CALLBACK EnumZBufferCallback(LPDDPIXELFORMAT pf, void* dst)
{
	con << "Z buffer format\n"
		<< pf
		<< "-------------------------------------\n";

	// If this is ANY type of depth-buffer, stop.
	if (pf->dwFlags == DDPF_ZBUFFER)
	{
		memcpy(dst, pf, sizeof(DDPIXELFORMAT));
 
		// Return with D3DENUMRET_CANCEL to end the search.
		return D3DENUMRET_CANCEL;
	}
 
	// Return with D3DENUMRET_OK to continue the search.
	return D3DENUMRET_OK;
}

//==========================================================================
//
//	EnumPixelFormatsCallback
//
//==========================================================================

static HRESULT CALLBACK EnumPixelFormatsCallback(LPDDPIXELFORMAT pf, void* dst)
{
	con << "Pixel format\n"
		<< pf
		<< "-------------------------------------\n";

	if ((pf->dwFlags == (DDPF_RGB|DDPF_ALPHAPIXELS)) &&
		(pf->dwRGBBitCount == 16))
	{
		memcpy(dst, pf, sizeof(DDPIXELFORMAT));
 
		// Return with D3DENUMRET_CANCEL to end the search.
		return D3DENUMRET_CANCEL;
	}
	con << "Flags " << pf->dwFlags << " need " << (DDPF_RGB|DDPF_ALPHAPIXELS) << endl;

	// Return with D3DENUMRET_OK to continue the search.
	return D3DENUMRET_OK;
}

//==========================================================================
//
//	GetBits
//
//==========================================================================

inline int GetBits(dword mask)
{
	int answer = 0;
	while (mask)
	{
		if (mask & 1)
		{
			answer++;
		}
		mask >>= 1;
	}
	return answer;
}

//==========================================================================
//
//	GetShift
//
//==========================================================================

inline int GetShift(dword mask)
{
	if (!mask)
		return 0;
	int answer = 0;
	while (!(mask & 1))
	{
		answer++;
		mask >>= 1;
	}
	return answer;
}

//==========================================================================
//
// 	TDirect3DDrawer::SetResolution
//
// 	Set up the video mode
//
//==========================================================================

bool TDirect3DDrawer::SetResolution(int Width, int Height, int BPP)
{
	if (!Width || !Height)
	{
		//	Set defaults
		Width = 640;
		Height = 480;
		BPP = 16;
	}

	if (BPP == 15) BPP = 16;

	if (BPP < 16)
	{
		//	True-color only
		return false;
	}

	con << "-------------------------------------\n";

	DDSURFACEDESC2	ddsd;

	//	Shut down current mode
	ReleaseTextures();
	SAFE_RELEASE(RenderDevice)
	SAFE_RELEASE(ZBuffer)
	SAFE_RELEASE(RenderSurface)
	SAFE_RELEASE(PrimarySurface)

	//	Set video mode
	if (DDraw->SetDisplayMode(Width, Height, BPP, 0, 0) != DD_OK)
		return false;

	//	Create primary surface
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	if (DDraw->CreateSurface(&ddsd, &PrimarySurface, NULL) != DD_OK)
		I_Error("I_SetResolution: Failed to create primary surface");

	Direct3D->EnumDevices(EnumDevicesCallback, NULL);

	GUID DeviceGUID;
	if (use_hal)
	{
		DeviceGUID = IID_IDirect3DHALDevice;
		SurfaceMemFlag = DDSCAPS_VIDEOMEMORY;
	}
	else
	{
		DeviceGUID = IID_IDirect3DRGBDevice;
		SurfaceMemFlag = DDSCAPS_SYSTEMMEMORY;
	}

	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_3DDEVICE;
    ddsd.dwWidth  = Width;
    ddsd.dwHeight = Height;
 
    //	Create the back buffer.
	if (FAILED(DDraw->CreateSurface(&ddsd, &RenderSurface, NULL)))
		return false;

    //	Create the z-buffer
    DDPIXELFORMAT ddpfZBuffer;
	memset(&ddpfZBuffer, 0, sizeof(ddpfZBuffer));
	Direct3D->EnumZBufferFormats(DeviceGUID,
		EnumZBufferCallback, (VOID*)&ddpfZBuffer);

	//	If the enumerated format is good (it should be), the dwSize member
	// will be properly initialized. Check this just in case.
	if (sizeof(DDPIXELFORMAT) != ddpfZBuffer.dwSize)
		return false;
 
	// Get z-buffer dimensions from the render target
	// Setup the surface desc for the z-buffer.
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER | SurfaceMemFlag;
	ddsd.dwWidth = Width;
	ddsd.dwHeight = Height;
	memcpy(&ddsd.ddpfPixelFormat, &ddpfZBuffer, sizeof(DDPIXELFORMAT));

	// Create the depth-buffer.
	if (FAILED(DDraw->CreateSurface(&ddsd, &ZBuffer, NULL)))
        return false;

	// Attach the z-buffer to the back buffer.
	if (FAILED(RenderSurface->AddAttachedSurface(ZBuffer)))
		return false;

	//	Create rendering device
	if (FAILED(Direct3D->CreateDevice(DeviceGUID,
		RenderSurface, &RenderDevice)))
	{
		return false;
    }

	D3DDEVICEDESC7	DeviceDesc;
	RenderDevice->GetCaps(&DeviceDesc);
	HavesAlphaTest = (DeviceDesc.dpcTriCaps.dwAlphaCmpCaps & D3DPCMPCAPS_GREATEREQUAL) != 0;
	// needed?
	//minTexSize = MAX(DeviceDesc.dwMinTextureWidth, DeviceDesc.dwMinTextureHeight);
	maxTexSize = MAX(DeviceDesc.dwMaxTextureWidth, DeviceDesc.dwMaxTextureHeight);

	memset(&PixelFormat, 0, sizeof(PixelFormat));
 	RenderDevice->EnumTextureFormats(EnumPixelFormatsCallback, &PixelFormat);
	if (PixelFormat.dwSize != sizeof(DDPIXELFORMAT))
		return false;
	abit = PixelFormat.dwRGBAlphaBitMask;
	rshift = GetShift(PixelFormat.dwRBitMask);
	gshift = GetShift(PixelFormat.dwGBitMask);
	bshift = GetShift(PixelFormat.dwBBitMask);

	//	Set screen params
	ScreenWidth = Width;
	ScreenHeight = Height;
	ScreenPitch = Width;
	ScreenBPP = BPP;

//	Sys_HighFPPrecision();

	return true;
}

//==========================================================================
//
//	TDirect3DDrawer::InitResolution
//
//==========================================================================

void TDirect3DDrawer::InitResolution(void)
{
	RenderDevice->SetTransform(D3DTRANSFORMSTATE_WORLD, &IdentityMatrix);
 
	RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_GREATER);
	RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHAREF, 0);

	RenderDevice->SetRenderState(D3DRENDERSTATE_SRCBLEND, D3DBLEND_SRCALPHA);
	RenderDevice->SetRenderState(D3DRENDERSTATE_DESTBLEND, D3DBLEND_INVSRCALPHA);

	RenderDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);

	RenderDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_MODULATEALPHA);

	RenderDevice->SetRenderState(D3DRENDERSTATE_FOGTABLEMODE, D3DFOG_LINEAR);
	RenderDevice->SetRenderState(D3DRENDERSTATE_FOGCOLOR, 0x007f7f7f);
	RenderDevice->SetRenderState(D3DRENDERSTATE_FOGDENSITY, PassFloat(0.5));
	RenderDevice->SetRenderState(D3DRENDERSTATE_FOGSTART, PassFloat(0.5));
	RenderDevice->SetRenderState(D3DRENDERSTATE_FOGEND, PassFloat(0.8));
}

//==========================================================================
//
//	TDirect3DDrawer::StartUpdate
//
//==========================================================================

void TDirect3DDrawer::StartUpdate(void)
{
	// Check for lost surface
	if (RenderSurface->IsLost() != DD_OK)
	{
		RenderSurface->Restore();
	}

	//	Clear surface
	if (clear)
	{
		RenderDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0, 1.0, 0);
	}
	else
	{
		RenderDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER, 0, 1.0, 0);
	}

	//	Setup texture filtering
	if (tex_linear)
	{
		RenderDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTFG_LINEAR);
		RenderDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTFN_LINEAR);
	}
	else
	{
		RenderDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTFG_POINT);
		RenderDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTFN_POINT);
	}

	//	Dithering
	if (dither)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_DITHERENABLE, TRUE);
	}
	else
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_DITHERENABLE, FALSE);
	}

	// Begin the scene.
	RenderDevice->BeginScene();

	Setup2D();
}

//==========================================================================
//
//	TDirect3DDrawer::SetupView
//
//==========================================================================

void TDirect3DDrawer::SetupView(int x, int y, int width, int height, float fovx, float fovy)
{
	//	Setup viewport
	memset(&viewData, 0, sizeof(D3DVIEWPORT7));
	viewData.dwX = x;
	viewData.dwY = y;
	viewData.dwWidth  = width;
	viewData.dwHeight = height;
	viewData.dvMinZ = 0.0f;
	viewData.dvMaxZ = 1.0f;

	//	Setup projection
	memset(&matProj, 0, sizeof(D3DMATRIX));
	matProj(0, 0) = 1 / fovx;
	matProj(1, 1) = 1 / fovy;
	matProj(2, 2) = 1;
	matProj(2, 3) = 1;
	matProj(3, 2) = -0.1;
}

//==========================================================================
//
//	TDirect3DDrawer::SetupFrame
//
//==========================================================================

void TDirect3DDrawer::SetupFrame(void)
{
	if (viewwidth != ScreenWidth)
	{
		R_DrawViewBorder();
	}

    RenderDevice->SetViewport(&viewData);
	RenderDevice->SetTransform(D3DTRANSFORMSTATE_PROJECTION, &matProj);

	// The view matrix defines the position and orientation of the camera.
	D3DMATRIX matView;
	matView(0, 0) = viewright.x;
	matView(1, 0) = viewright.y;
	matView(2, 0) = viewright.z;
	matView(3, 0) = -DotProduct(vieworg, viewright);
	matView(0, 1) = viewup.x;
	matView(1, 1) = viewup.y;
	matView(2, 1) = viewup.z;
	matView(3, 1) = -DotProduct(vieworg, viewup);
	matView(0, 2) = viewforward.x;
	matView(1, 2) = viewforward.y;
	matView(2, 2) = viewforward.z;
	matView(3, 2) = -DotProduct(vieworg, viewforward);
	matView(0, 3) = 0;
	matView(1, 3) = 0;
	matView(2, 3) = 0;
	matView(3, 3) = 1;
	RenderDevice->SetTransform(D3DTRANSFORMSTATE_VIEW, &matView);

 	if (use_zlight)
	{
		RenderDevice->SetRenderState(D3DRENDERSTATE_SHADEMODE, D3DSHADE_GOURAUD);
	}
	RenderDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, TRUE);
	RenderDevice->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE);
	RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	RenderDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, FALSE);
}

//==========================================================================
//
//	TDirect3DDrawer::WorldDrawing
//
//==========================================================================

void TDirect3DDrawer::WorldDrawing(void)
{
}

//==========================================================================
//
//	TDirect3DDrawer::EndView
//
//==========================================================================

void TDirect3DDrawer::EndView(void)
{
	Setup2D();

	cl.cshifts[7] = cl.prev_cshifts[7];
	for (int i = 0; i < NUM_CSHIFTS; i++)
	{
		if (!cl.cshifts[i])
		{
			continue;
		}

		D3DLVERTEX	dv[4];

		RenderDevice->SetTexture(0, NULL);
		if (HavesAlphaTest)
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
		}

		dv[0] = D3DLVERTEX(D3DVECTOR(0, 0, 0), cl.cshifts[i], 0, 0, 0);
		dv[1] = D3DLVERTEX(D3DVECTOR(ScreenWidth, 0, 0), cl.cshifts[i], 0, 0, 0);
		dv[2] = D3DLVERTEX(D3DVECTOR(ScreenWidth, ScreenHeight, 0), cl.cshifts[i], 0, 0, 0);
		dv[3] = D3DLVERTEX(D3DVECTOR(0, ScreenHeight, 0), cl.cshifts[i], 0, 0, 0);

		RenderDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_LVERTEX, dv, 4, 0);

		if (HavesAlphaTest)
		{
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
			RenderDevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		}
	}
}

//==========================================================================
//
//	TDirect3DDrawer::Update
//
// 	Blit to the screen / Flip surfaces
//
//==========================================================================

void TDirect3DDrawer::Update(void)
{
	// End the scene.
	RenderDevice->EndScene();

	// Check for lost surface
	if (PrimarySurface->IsLost() != DD_OK)
	{
		PrimarySurface->Restore();
	}

	PrimarySurface->Blt(NULL, RenderSurface, NULL, DDBLT_WAIT, NULL);
}

//==========================================================================
//
// 	TDirect3DDrawer::Shutdown
//
//	Close the graphics
//
//==========================================================================

void TDirect3DDrawer::Shutdown(void)
{
	ReleaseTextures();
	SAFE_RELEASE(RenderDevice)
	SAFE_RELEASE(Direct3D)

	SAFE_RELEASE(ZBuffer)
	SAFE_RELEASE(RenderSurface)
	SAFE_RELEASE(PrimarySurface)
	SAFE_RELEASE(DDraw)
}

//==========================================================================
//
//	TDirect3DDrawer::ReadScreen
//
//==========================================================================

void *TDirect3DDrawer::ReadScreen(int *bpp, bool *bot2top)
{
	DDSURFACEDESC2	ddsd;

	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_LPSURFACE;
	if (RenderSurface->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL) != DD_OK)
		I_Error("V_Update: Failed to lock screen");
	void *dst = Z_Malloc(ScreenPitch * ScreenHeight * PixelBytes);
	memcpy(dst, ddsd.lpSurface, ScreenPitch * ScreenHeight * PixelBytes);
	RenderSurface->Unlock(NULL);
	*bpp = ScreenBPP;
	*bot2top = false;
	return dst;
}

//==========================================================================
//
//	TDirect3DDrawer::InitWipe
//
//==========================================================================

bool TDirect3DDrawer::InitWipe(void)
{
	return false;
}

//==========================================================================
//
//	TDirect3DDrawer::DoWipe
//
//==========================================================================

void TDirect3DDrawer::DoWipe(int)
{
}

//==========================================================================
//
//	TDirect3DDrawer::SetPalette
//
//==========================================================================

void TDirect3DDrawer::SetPalette(int pnum)
{
	rgb_t *pal = (rgb_t*)W_CacheLumpName("playpal", PU_CACHE);
	for (int i = 0; i < 256; i++)
	{
		pal8_to16[i] = MakeCol16(pal[i].r, pal[i].g, pal[i].b);
	}

	pal += 256 * pnum;
	int cmax = MAX(MAX(pal[0].r, pal[0].g), pal[0].b);
	if (!cmax)
	{
		cl.cshifts[7] = 0;
	}
	else
	{
		cl.cshifts[7] = (cmax << 24) | ((255 * pal[0].r / cmax) << 16) |
			((255 * pal[0].g / cmax) << 8) | (255 * pal[0].b / cmax);
	}
	cl.prev_cshifts[7] = cl.cshifts[7];
}

//==========================================================================
//
//
//
//==========================================================================

ostream &operator << (ostream &str, const D3DPRIMCAPS pc)
{
	str << "Supported alpaha tests:";
	if (pc.dwAlphaCmpCaps & D3DPCMPCAPS_ALWAYS)
	{
		str << " always";
	}
	if (pc.dwAlphaCmpCaps & D3DPCMPCAPS_EQUAL)
	{
		str << " equal";
	}
	if (pc.dwAlphaCmpCaps & D3DPCMPCAPS_GREATER)
	{
		str << " greater";
	}
	if (pc.dwAlphaCmpCaps & D3DPCMPCAPS_GREATEREQUAL)
	{
		str << " greater_equal";
	}
	if (pc.dwAlphaCmpCaps & D3DPCMPCAPS_LESS)
	{
		str << " less";
	}
	if (pc.dwAlphaCmpCaps & D3DPCMPCAPS_LESSEQUAL)
	{
		str << " less_equal";
	}
	if (pc.dwAlphaCmpCaps & D3DPCMPCAPS_NEVER)
	{
		str << " never";
	}
	if (pc.dwAlphaCmpCaps & D3DPCMPCAPS_NOTEQUAL)
	{
		str << " not_equal";
	}
	str << endl;

	return str;
}

//==========================================================================
//
//
//
//==========================================================================

ostream &operator << (ostream &str, const LPD3DDEVICEDESC7 dd)
{
	str << "Device caps:\n";
	if (dd->dwDevCaps & D3DDEVCAPS_FLOATTLVERTEX)
	{
		str << "- accepts floating point for post-transform vertex data.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_SORTINCREASINGZ)
	{
		str << "- needs data sorted for increasing depth.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_SORTDECREASINGZ)
	{
		str << "- needs data sorted for decreasing depth.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_SORTEXACT)
	{
		str << "- needs data sorted exactly.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_EXECUTESYSTEMMEMORY)
	{
		str << "- can use execute buffers from system memory.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_EXECUTEVIDEOMEMORY)
	{
		str << "- can use execute buffer from video memory.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_TLVERTEXSYSTEMMEMORY)
	{
		str << "- can use buffers from system memory for transformed and lit vertices.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_TLVERTEXVIDEOMEMORY)
	{
		str << "- can use buffers from video memory for transformed and lit vertices.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_TEXTURESYSTEMMEMORY)
	{
		str << "- can retrieve textures from system memory.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_TEXTUREVIDEOMEMORY)
	{
		str << "- can retrieve textures from device memory.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_DRAWPRIMTLVERTEX)
	{
		str << "- exports a DrawPrimitive-aware HAL.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_CANRENDERAFTERFLIP)
	{
		str << "- can queue rendering commands after a page flip.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_TEXTURENONLOCALVIDMEM)
	{
		str << "- can retrieve textures from non-local video (AGP) memory.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_DRAWPRIMITIVES2)
	{
		str << "- can support DrawPrimitives2.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_SEPARATETEXTUREMEMORIES)
	{
		str << "- uses discrete texture memory pools for each stage.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_DRAWPRIMITIVES2EX)
	{
		str << "- can support Extended DrawPrimitives2 i.e. DX7 compliant driver.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT)
	{
		str << "- supports transformation and lighting in hardware.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_CANBLTSYSTONONLOCAL)
	{
		str << "- supports blits from system-memory textures to non-local video-memory textures.\n";
	}
	if (dd->dwDevCaps & D3DDEVCAPS_HWRASTERIZATION)
	{
		str << "- has HW acceleration for rasterization.\n";
	}
	str << endl;

	str << "Line caps:\n" << dd->dpcLineCaps;
	str << "Triangle caps:\n" << dd->dpcTriCaps;
 	str << endl;

	str << "Rendering bit depth: ";
 	if (dd->dwDeviceRenderBitDepth & DDBD_8) str << "8 ";
 	if (dd->dwDeviceRenderBitDepth & DDBD_16) str << "16 ";
 	if (dd->dwDeviceRenderBitDepth & DDBD_24) str << "24 ";
 	if (dd->dwDeviceRenderBitDepth & DDBD_32) str << "32 ";
 	str << endl;
	str << "Z-Buffer bit depth: ";
 	if (dd->dwDeviceZBufferBitDepth & DDBD_8) str << "8 ";
 	if (dd->dwDeviceZBufferBitDepth & DDBD_16) str << "16 ";
 	if (dd->dwDeviceZBufferBitDepth & DDBD_24) str << "24 ";
 	if (dd->dwDeviceZBufferBitDepth & DDBD_32) str << "32 ";
 	str << endl;
	str << endl;

	str << "Min texture width: " << dd->dwMinTextureWidth << endl;
	str << "Min texture height: " << dd->dwMinTextureHeight << endl;
	str << "Max texture width: " << dd->dwMaxTextureWidth << endl;
	str << "Max texture height: " << dd->dwMaxTextureHeight << endl;
	str << endl;

	str << "Max texture repeat: " << dd->dwMaxTextureRepeat << endl;
	str << "Max texture aspect ratio: " << dd->dwMaxTextureAspectRatio << endl;
	str << "Max anisotropy: " << dd->dwMaxAnisotropy << endl;
	str << endl;

	str << "Guard band left: " << dd->dvGuardBandLeft << endl;
	str << "Guard band top: " << dd->dvGuardBandTop << endl;
	str << "Guard band right: " << dd->dvGuardBandRight << endl;
	str << "Guard band bottom: " << dd->dvGuardBandBottom << endl;
	str << endl;

	str << "Extents adjust: " << dd->dvExtentsAdjust << endl;
	str << endl;

	str << "Stencil caps: ";
	if (dd->dwStencilCaps & D3DSTENCILCAPS_DECR)
	{
		str << "DECR ";
	}
	if (dd->dwStencilCaps & D3DSTENCILCAPS_DECRSAT)
	{
		str << "DECRSAT	";
	}
	if (dd->dwStencilCaps & D3DSTENCILCAPS_INCR)
	{
		str << "INCR ";
	}
	if (dd->dwStencilCaps & D3DSTENCILCAPS_INCRSAT)
	{
		str << "INCRSAT ";
	}
	if (dd->dwStencilCaps & D3DSTENCILCAPS_INVERT)
	{
		str << "INVERT ";
	}
	if (dd->dwStencilCaps & D3DSTENCILCAPS_KEEP)
	{
		str << "KEEP ";
	}
	if (dd->dwStencilCaps & D3DSTENCILCAPS_REPLACE)
	{
		str << "REPLACE ";
	}
	if (dd->dwStencilCaps & D3DSTENCILCAPS_ZERO)
	{
		str << "ZERO ";
	}
	str << endl;

	str << "Flexible vertex format capabilities:\n";
	if (dd->dwFVFCaps & D3DFVFCAPS_DONOTSTRIPELEMENTS)
	{
		str << "- prefers that vertex elements not be stripped.\n";
	}
	str << "- total " << (dd->dwFVFCaps & D3DFVFCAPS_TEXCOORDCOUNTMASK) << " texture coordinate sets\n";

	str << "Supported texture operations: ";
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_ADD)
	{
		str << "ADD ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_ADDSIGNED)
	{
		str << "ADDSIGNED ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_ADDSIGNED2X)
	{
		str << "ADDSIGNED2X ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_ADDSMOOTH)
	{
		str << "ADDSMOOTH ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_BLENDCURRENTALPHA)
	{
		str << "BLENDCURRENTALPHA ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_BLENDDIFFUSEALPHA)
	{
		str << "BLENDDIFFUSEALPHA ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_BLENDFACTORALPHA)
	{
		str << "BLENDFACTORALPHA ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_BLENDTEXTUREALPHA)
	{
		str << "BLENDTEXTUREALPHA ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_BLENDTEXTUREALPHAPM)
	{
		str << "BLENDTEXTUREALPHAPM ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_BUMPENVMAP)
	{
		str << "BUMPENVMAP ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_BUMPENVMAPLUMINANCE)
	{
		str << "BUMPENVMAPLUMINANCE ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_DISABLE)
	{
		str << "DISABLE ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_DOTPRODUCT3)
	{
		str << "DOTPRODUCT3 ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_MODULATE)
	{
		str << "MODULATE ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_MODULATE2X)
	{
		str << "MODULATE2X ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_MODULATE4X)
	{
		str << "MODULATE4X ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_MODULATEALPHA_ADDCOLOR)
	{
		str << "MODULATEALPHA_ADDCOLOR ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_MODULATECOLOR_ADDALPHA)
	{
		str << "MODULATEALPHA_ADDCOLOR ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_MODULATEINVALPHA_ADDCOLOR)
	{
		str << "MODULATEINVALPHA_ADDCOLOR ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_MODULATEINVCOLOR_ADDALPHA)
	{
		str << "MODULATEINVCOLOR_ADDALPHA ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_PREMODULATE)
	{
		str << "PREMODULATE ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_SELECTARG1)
	{
		str << "SELECTARG1 ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_SELECTARG2)
	{
		str << "SELECTARG2 ";
	}
	if (dd->dwTextureOpCaps & D3DTEXOPCAPS_SUBTRACT)
	{
		str << "SUBTRACT ";
	}
	str << endl;
	str << "Max texture blend stages: " << dd->wMaxTextureBlendStages << endl;
	str << "Max simultaneous textures: " << dd->wMaxSimultaneousTextures << endl;
	str << "Max active lights: " << dd->dwMaxActiveLights << endl;
	str << "Max vertex W: " << dd->dvMaxVertexW << endl;
	str << "Max user clip planes: " << dd->wMaxUserClipPlanes << endl;
	str << "Max vertex blend matrices: " << dd->wMaxVertexBlendMatrices << endl;

/*
    GUID        deviceGUID;

    DWORD       dwVertexProcessingCaps;
*/
	return str;
}

//==========================================================================
//
//
//
//==========================================================================

ostream &operator << (ostream &str, const LPDDPIXELFORMAT pf)
{
	if (pf->dwFlags & DDPF_ALPHAPIXELS)
		str << "- has alpha channel." << endl;
	if (pf->dwFlags & DDPF_ALPHA)
		str << "- Alpha-only surface." << endl;
	if (pf->dwFlags & DDPF_FOURCC)
		str << "- contains a FOURCC code describing a non-RGB pixel format." << endl;
	if (pf->dwFlags & DDPF_PALETTEINDEXED1)
		str << "- 1-bit color indexed." << endl;
	if (pf->dwFlags & DDPF_PALETTEINDEXED2)
		str << "- 2-bit color indexed." << endl;
	if (pf->dwFlags & DDPF_PALETTEINDEXED4)
		str << "- 4-bit color indexed." << endl;
	if (pf->dwFlags & DDPF_PALETTEINDEXED8)
		str << "- 8-bit color indexed." << endl;
	if (pf->dwFlags & DDPF_PALETTEINDEXEDTO8)
		str << "- indexed to an 8-bit palette." << endl;
	if (pf->dwFlags & DDPF_RGB)
		str << "- RGB surface." << endl;
	if (pf->dwFlags & DDPF_COMPRESSED)
		str << "- will accept pixel data in the specified format and compress it during the write operation." << endl;
	if (pf->dwFlags & DDPF_RGBTOYUV)
		str << "- translates RGB data YUV data." << endl;
	if (pf->dwFlags & DDPF_YUV)
		str << "- YUV surface." << endl;
	if (pf->dwFlags & DDPF_ZBUFFER)
		str << "- z-buffer surface." << endl;
	if (pf->dwFlags & DDPF_ZPIXELS)
		str << "- contains z information in the pixels." << endl;
	if (pf->dwFlags & DDPF_STENCILBUFFER)
		str << "- contains stencil information along with Z." << endl;
	if (pf->dwFlags & DDPF_ALPHAPREMULT)
		str << "- uses the premultiplied alpha format." << endl;
	if (pf->dwFlags & DDPF_LUMINANCE)
		str << "- luminance-only or luminance-alpha surface." << endl;
	if (pf->dwFlags & DDPF_BUMPLUMINANCE)
		str << "- luminance off bumpmap surface." << endl;
	if (pf->dwFlags & DDPF_BUMPDUDV)
		str << "- has Bump-map." << endl;
	str << "- Bit depth: " << pf->dwRGBBitCount << endl;
	str << "- Mask for red bits: " << hex << pf->dwRBitMask << dec << endl;
	str << "- Mask for green bits: " << hex << pf->dwGBitMask << dec << endl;
	str << "- Mask for blue bits: " << hex << pf->dwBBitMask << dec << endl;
	str << "- Mask for alpha/Z channel: " << hex << pf->dwRGBAlphaBitMask << dec << endl;

	return str;
}


