//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern byte		gammatable[5][256];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int				black_color;

//
//	Colormaps
//
byte			*colormaps;	// Standard colormap
byte			*fadetable;	// Current level's colormap
word			*fadetable16;
byte			*fogmap;
word			*fogmap16;

//
//	Translucency tables
//
byte 			*tinttables[5];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int ft_size = 0;

static rgb_t	host_basepal[256];

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	CalcCol16Table
//
//==========================================================================

static void CalcCol16Table(void)
{
	rgb_t *pal = (rgb_t*)W_CacheLumpName("playpal", PU_CACHE);
	byte *gt = gammatable[usegamma];
	for (int i = 0; i < 256; i++)
	{
		pal8_to16[i] = MakeCol(gt[pal[i].r], gt[pal[i].g], gt[pal[i].b]);
	}
}

//==========================================================================
//
//	CalcFadetable16
//
//==========================================================================

static void CalcFadetable16(rgb_t *pal)
{
	byte *gt = gammatable[usegamma];
	int i = 0;
	for (int tn = 0; tn < 32; tn++)
	{
		int colm = 32 - tn;
		int fog = 128 * tn / 32;
		for (int ci = 0; ci < 256; ci++, i++)
		{
			fadetable16[i] = MakeCol(
				gt[pal[ci].r] * colm / 32,
				gt[pal[ci].g] * colm / 32,
				gt[pal[ci].b] * colm / 32);
			fogmap16[i] = MakeCol(
				gt[pal[ci].r] * colm / 32 + fog,
				gt[pal[ci].g] * colm / 32 + fog,
				gt[pal[ci].b] * colm / 32 + fog);
			//	For 16 bit we use color 0 as transparent
			if (!fadetable16[i])
			{
				fadetable16[i] = 1;
			}
			if (!fogmap16[i])
			{
				fogmap16[i] = 1;
			}
		}
	}
	for (; i < ft_size; i++)
	{
		fadetable16[i] = pal8_to16[fadetable[i]];
		fogmap16[i] = pal8_to16[fadetable[i]];
		//	For 16 bit we use color 0 as transparent
		if (!fadetable16[i])
		{
			fadetable16[i] = 1;
		}
		if (!fogmap16[i])
		{
			fogmap16[i] = 1;
		}
	}

	D_FlushCaches();
}

//==========================================================================
//
//	TSoftwareDrawer::SetPalette
//
//==========================================================================

void TSoftwareDrawer::SetPalette(int num)
{
	if (num < 0 || num >= 32)
	{
		cond << "Invalid palette num " << num << endl;
		num = 0;
	}
	byte *pal = (byte*)W_CacheLumpName("playpal", PU_CACHE) + num * 768;
	if (ScreenBPP == 8)
	{
		SetPalette8(pal);
	}
	else
	{
		CalcCol16Table();
		CalcFadetable16((rgb_t*)pal);
	}

	//	Save palette
	memcpy(host_basepal, pal, 768);
	//	Must recalculate any cshifts
	memset(cl.prev_cshifts, 0, sizeof(cl.prev_cshifts));
}

//==========================================================================
//
//	InitColormaps
//
//==========================================================================

static void InitColormaps(void)
{
	int i, lump;

	//	We use color 0 as transparent color, so we must find an alternate
	// index for black color. In Doom, Heretic and Strife there is another
	// black color, in Hexen it's almost black.
	//	I think that originaly Doom uses color 255 as transparent color, but
	// utilites created by others uses the alternate black color and these
	// graphics can contain pixels of color 255.
	//	Heretic and Hexen also uses color 255 as transparent, even more - in
	// colormaps it's maped to color 0. Posibly this can cause problems with
	// modified graphics.
	//	Strife uses color 0 as transparent. I already had problems with fact
	// that color 255 is normal color, now there shouldn't be any problems.
	rgb_t *pal = (rgb_t*)W_CacheLumpName("playpal", PU_CACHE);
	int best_dist = 0x10000;
	for (i = 1; i < 256; i++)
	{
		int dist = pal[i].r * pal[i].r + pal[i].g * pal[i].g + pal[i].b * pal[i].b;
		if (dist < best_dist)
		{
			black_color = i;
			best_dist = dist;
		}
	}

    // Load in the light tables,
	lump = W_GetNumForName("COLORMAP");
	ft_size = W_LumpLength(lump);
    colormaps = (byte*)W_CacheLumpNum(lump, PU_STATIC);
    fadetable = colormaps;
	fadetable16 = (word*)Z_Malloc(ft_size * 2);

	//	Remap color 0 to alternate balck color
	for (i = 0; i < ft_size; i++)
	{
		if (!colormaps[i])
		{
			colormaps[i] = black_color;
		}
	}

	lump = W_CheckNumForName("FOGMAP");
	if (lump < 0)
	{
		fogmap = fadetable;
	}
	else
	{
		fogmap = (byte*)W_CacheLumpNum(lump, PU_STATIC);
		//	Remap also fogmap
		for (i = 0; i < ft_size; i++)
		{
			if (!fogmap[i])
			{
				fogmap[i] = black_color;
			}
		}
	}
	fogmap16 = (word*)Z_Malloc(ft_size * 2);
}

//==========================================================================
//
//	InitTranslucencyTables
//
//==========================================================================

static void InitTranslucencyTables(void)
{
    tinttables[0] = (byte*)W_CacheLumpName("TRANSP10", PU_STATIC);
    tinttables[1] = (byte*)W_CacheLumpName("TRANSP20", PU_STATIC);
    tinttables[2] = (byte*)W_CacheLumpName("TRANSP30", PU_STATIC);
    tinttables[3] = (byte*)W_CacheLumpName("TRANSP40", PU_STATIC);
    tinttables[4] = (byte*)W_CacheLumpName("TRANSP50", PU_STATIC);
}

//==========================================================================
//
//	TSoftwareDrawer::InitData
//
//==========================================================================

void TSoftwareDrawer::InitData(void)
{
    InitLightTables();
    InitColormaps();
	InitTranslucencyTables();
}

//==========================================================================
//
//	TSoftwareDrawer::UpdatePalette
//
//==========================================================================

void TSoftwareDrawer::UpdatePalette(void)
{
	int		i, j;
	bool	newshifts;
	byte	*basepal, *newpal;
	byte	pal[768];
	int		r,g,b;
	int		dstr, dstg, dstb, perc;

	newshifts = false;
	
	for (i = 0; i < NUM_CSHIFTS; i++)
	{
		if (cl.cshifts[i] != cl.prev_cshifts[i])
		{
			newshifts = true;
			cl.prev_cshifts[i] = cl.cshifts[i];
		}
	}
	
	if (!newshifts)
	{
		return;
	}
			
	basepal = (byte*)host_basepal;
	newpal = pal;
	
	for (i = 0; i < 256; i++)
	{
		r = basepal[0];
		g = basepal[1];
		b = basepal[2];
		basepal += 3;
	
		for (j = 0; j < NUM_CSHIFTS; j++)
		{
			perc = (cl.cshifts[j] >> 24) & 0xff;
			dstr = (cl.cshifts[j] >> 16) & 0xff;
			dstg = (cl.cshifts[j] >> 8) & 0xff;
			dstb = cl.cshifts[j] & 0xff;
			r += (perc * (dstr - r)) >> 8;
			g += (perc * (dstg - g)) >> 8;
			b += (perc * (dstb - b)) >> 8;
		}
		
		newpal[0] = r;
		newpal[1] = g;
		newpal[2] = b;
		newpal += 3;
	}

	if (ScreenBPP == 8)
	{
		SetPalette8(pal);
	}
	else
	{
		CalcCol16Table();
		CalcFadetable16((rgb_t*)pal);
	}
}

