//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**
//**	Local header for software drawer
//**
//**************************************************************************

#ifndef _D_LOCAL_H
#define _D_LOCAL_H

// HEADER FILES ------------------------------------------------------------

#include "gamedefs.h"
#include "cl_local.h"
#include "r_shared.h"

// MACROS ------------------------------------------------------------------

#define DS_SPAN_LIST_END	-128

// TYPES -------------------------------------------------------------------

struct sspan_t
{
	int		u;
	int		v;
	int		count;
};

typedef void (*spanfunc_t)(void);
typedef void (*spritespanfunc_t)(sspan_t*);

typedef void (*picspanfunc_t)(int, int, fixed_t, fixed_t, fixed_t, int, byte*);

class TSoftwareDrawer : public TDrawer
{
 public:
	TSoftwareDrawer(void);
	void Init(void);
	void InitData(void);
	bool SetResolution(int, int, int);
	void InitResolution(void);
	void SetPalette(int);
	void SetPalette8(byte*);
	void StartUpdate(void);
	void Update(void);
	void Shutdown(void);
	void* ReadScreen(int*, bool*);

	//	Screen wipes
	bool InitWipe(void);
	void DoWipe(int);

	//	Rendering stuff
	void SetupView(int, int, int, int, float, float);
	void SetupFrame(void);
	void WorldDrawing(void);
	void EndView(void);

	//	Texture stuff
	void InitTextures(void);
	void SetTexture(int);
	void SetSkyTexture(int, bool);
	void SetFlat(int);
	void SetSpriteLump(int, int, int);

	//	Polygon drawing
	void DrawPolygon(draw_vertex_t *, int, int, int, bool);
	void DrawMaskedPolygon(draw_vertex_t *, int, int, int, int, bool);
	void DrawZPolygon(draw_vertex_t *, int, int);
	void DrawSkyPolygon(draw_vertex_t *, int, int, bool);
	void DrawSpritePolygon(draw_vertex_t *, int, int, int, bool);

	//	Drawing
	void DrawPic(float, float, float, float, float, float, float, float, int, int);
	void DrawPicShadow(float, float, float, float, float, float, float, float, int, int);
	void FillRectWithFlat(int, int, int, int, const char*);
	void ShadeRect(int, int, int, int, int);
	void DrawConsoleBackground(int);
	void DrawSpriteLump(int, int, int, int, boolean);

 private:
	void InitLightTables(void);
	void UpdatePalette(void);
};

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

extern "C" {

void D_DrawSpan8_8(void);
void D_DrawMaskedSpan_8(void);
void D_DrawFuzzSpan_8(void);
void D_DrawAltFuzzSpan_8(void);
void D_DrawSkySpan_8(void);
void D_DrawDoubleSkySpan_8(void);
void D_DrawSpriteSpan_8(sspan_t*);
void D_DrawFuzzSpriteSpan_8(sspan_t*);
void D_DrawAltFuzzSpriteSpan_8(sspan_t*);

void D_DrawSpan8_16(void);
void D_DrawMaskedSpan_16(void);
void D_DrawFuzzSpan_15(void);
void D_DrawFuzzSpan_16(void);
void D_DrawSkySpan_16(void);
void D_DrawDoubleSkySpan_16(void);
void D_DrawSpriteSpan_16(sspan_t*);
void D_DrawFuzzSpriteSpan_15(sspan_t*);
void D_DrawFuzzSpriteSpan_16(sspan_t*);

void D_DrawZSpan(void);

void D_DrawTriangle(draw_vertex_t *vb, int v1, int v2, int v3);

} // extern "C"

void TransformVector(const TVec &in, TVec &out);

void D_ClearPolys(void);
void D_DrawWorldPolygon(draw_vertex_t *verts, int count, int masked);
void D_DrawTranslucentWorldPolygon(draw_vertex_t *verts, int count);
void D_SpriteDrawPolygon(draw_vertex_t *verts, int count);
void D_FlushCaches(void);

// PUBLIC DATA DECLARATIONS ------------------------------------------------

extern int				ds_u;
extern int				ds_v;
extern int				ds_count;
extern float			d_zistepu;
extern float			d_zistepv;
extern float			d_ziorigin;
extern float			d_sdivzstepu;
extern float			d_tdivzstepu;
extern float			d_sdivzstepv;
extern float			d_tdivzstepv;
extern float			d_sdivzorigin;
extern float			d_tdivzorigin;
extern fixed_t			sadjust;
extern fixed_t			tadjust;
extern fixed_t			bbextents;
extern fixed_t			bbextentt;
extern fixed_t			ds_l;
extern fixed_t			ds_lstepu;
extern byte*			ds_source;
extern byte*			ds_colormap;
extern word*			ds_colormap16;
extern int				ds_tex_width;
extern byte*			ds_transluc;
extern int				ds_transluc16;

extern int				ylookup[MAXSCREENHEIGHT];

extern spanfunc_t		spanfunc;
extern spritespanfunc_t	spritespanfunc;

extern spanfunc_t		D_DrawSpan;
extern spanfunc_t		D_DrawMaskedSpan;
extern spanfunc_t		D_DrawFuzzSpan;
extern spanfunc_t		D_DrawAltFuzzSpan;
extern spanfunc_t		D_DrawSkySpan;
extern spanfunc_t		D_DrawDoubleSkySpan;
extern spritespanfunc_t	D_DrawSpriteSpan;
extern spritespanfunc_t	D_DrawFuzzSpriteSpan;
extern spritespanfunc_t	D_DrawAltFuzzSpriteSpan;

extern float			centerxfrac;
extern float			centeryfrac;
extern float			xprojection;
extern float			yprojection;

extern byte				*colormaps;
extern byte				*fadetable;
extern word				*fadetable16;
extern byte				*fogmap;
extern word				*fogmap16;

extern byte				*tinttables[5];

extern byte				*scrn;
extern word				*scrn16;
extern int				usegamma;

extern word				*zbuffer;	//	Allocated by video driver

extern fixed_t			ScaleX;
extern fixed_t			ScaleY;
extern fixed_t			ScaleXI;
extern fixed_t			ScaleYI;

extern word				pal8_to16[256];

extern int				rshift;
extern int				gshift;
extern int				bshift;

extern byte				*fixedcolormap;
extern word				*fixedcolormap16;

extern int				black_color;

//==========================================================================
//
//	Inlines for 15bpp
//
//==========================================================================

inline dword MakeCol15(byte r, byte g, byte b)
{
	return	((r >> 3) << rshift) |
			((g >> 3) << gshift) |
			((b >> 3) << bshift);
}

inline byte GetCol15R(dword col)
{
	return ((col >> rshift) << 3) & 0xff;
}

inline byte GetCol15G(dword col)
{
	return ((col >> gshift) << 3) & 0xff;
}

inline byte GetCol15B(dword col)
{
	return ((col >> bshift) << 3) & 0xff;
}

//==========================================================================
//
//	Inlines for 16bpp
//
//==========================================================================

inline dword MakeCol16(byte r, byte g, byte b)
{
	return	((r >> 3) << rshift) |
			((g >> 2) << gshift) |
			((b >> 3) << bshift);
}

inline byte GetCol16R(dword col)
{
	return ((col >> rshift) << 3) & 0xff;
}

inline byte GetCol16G(dword col)
{
	return ((col >> gshift) << 2) & 0xff;
}

inline byte GetCol16B(dword col)
{
	return ((col >> bshift) << 3) & 0xff;
}

//==========================================================================
//
//	Inlines for 15 and 16 bit colors
//
//==========================================================================

inline dword MakeCol(byte r, byte g, byte b)
{
	switch (ScreenBPP)
	{
	 case 15:
		return MakeCol15(r, g, b);
	 case 16:
		return MakeCol16(r, g, b);
	 default:
		return 0;
	}
}

inline byte GetColR(dword col)
{
	switch (ScreenBPP)
	{
	 case 15:
		return GetCol15R(col);
	 case 16:
		return GetCol16R(col);
	 default:
		return 0;
	}
}

inline byte GetColG(dword col)
{
	switch (ScreenBPP)
	{
	 case 15:
		return GetCol15G(col);
	 case 16:
		return GetCol16G(col);
	 default:
		return 0;
	}
}

inline byte GetColB(dword col)
{
	switch (ScreenBPP)
	{
	 case 15:
		return GetCol15B(col);
	 case 16:
		return GetCol16B(col);
	 default:
		return 0;
	}
}

#endif

