//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

byte	*scrn;
word	*scrn16;
rgb_t	*scrn24;

byte	*fixedcolormap;
word	*fixedcolormap16;
rgb_t	*fixedcolormap24;

word	pal8_to16[256];
dword	pal2rgb[256];

int		rshift = 11;
int		gshift = 5;
int		bshift = 0;

// Scale values to draw VGA images
fixed_t	ScaleX;
fixed_t	ScaleY;
fixed_t	ScaleXI;
fixed_t	ScaleYI;

int				ylookup[MAXSCREENHEIGHT];

word			*zbuffer = NULL;

spanfunc_t			D_DrawSpan;
spanfunc_t			D_DrawMaskedSpan;
spanfunc_t			D_DrawFuzzSpan;
spanfunc_t			D_DrawAltFuzzSpan;
spanfunc_t			D_DrawSkySpan;
spanfunc_t			D_DrawDoubleSkySpan;
spritespanfunc_t	D_DrawSpriteSpan;
spritespanfunc_t	D_DrawFuzzSpriteSpan;
spritespanfunc_t	D_DrawAltFuzzSpriteSpan;

spanfunc_t			spanfunc;
spritespanfunc_t	spritespanfunc;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static TSoftwareDrawer		SoftwareDrawer;

static TCvarI			d_subdiv("d_subdiv", "1", CVAR_ARCHIVE);
static TCvarI			d_smooth("d_smooth", "0", CVAR_ARCHIVE);

static byte*			r_backscreen = NULL;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	TSoftwareDrawer::TSoftwareDrawer
//
//==========================================================================

TSoftwareDrawer::TSoftwareDrawer(void)
{
	_SoftwareDrawer = this;
}

//==========================================================================
//
//	TSoftwareDrawer::InitResolution
//
//	Calculate image scaling
//
//==========================================================================

void TSoftwareDrawer::InitResolution(void)
{
	scrn16 = (word*)scrn;

	if (ScreenBPP == 8)
	{
		D_DrawSpan = D_DrawSpan8_8;
		D_DrawMaskedSpan = D_DrawMaskedSpan_8;
		D_DrawFuzzSpan = D_DrawFuzzSpan_8;
		D_DrawAltFuzzSpan = D_DrawAltFuzzSpan_8;
		D_DrawSkySpan = D_DrawSkySpan_8;
		D_DrawDoubleSkySpan = D_DrawDoubleSkySpan_8;
		D_DrawSpriteSpan = D_DrawSpriteSpan_8;
		D_DrawFuzzSpriteSpan = D_DrawFuzzSpriteSpan_8;
		D_DrawAltFuzzSpriteSpan = D_DrawAltFuzzSpriteSpan_8;
	}
	else if (ScreenBPP == 15)
	{
		D_DrawSpan = D_DrawSpan8_16;
		D_DrawMaskedSpan = D_DrawMaskedSpan_16;
		D_DrawFuzzSpan = D_DrawFuzzSpan_15;
		D_DrawAltFuzzSpan = D_DrawFuzzSpan_15;
		D_DrawSkySpan = D_DrawSkySpan_16;
		D_DrawDoubleSkySpan = D_DrawDoubleSkySpan_16;
		D_DrawSpriteSpan = D_DrawSpriteSpan_16;
		D_DrawFuzzSpriteSpan = D_DrawFuzzSpriteSpan_15;
		D_DrawAltFuzzSpriteSpan = D_DrawFuzzSpriteSpan_15;
	}
	else if (ScreenBPP == 16)
	{
		D_DrawSpan = D_DrawSpan8_16;
		D_DrawMaskedSpan = D_DrawMaskedSpan_16;
		D_DrawFuzzSpan = D_DrawFuzzSpan_16;
		D_DrawAltFuzzSpan = D_DrawFuzzSpan_16;
		D_DrawSkySpan = D_DrawSkySpan_16;
		D_DrawDoubleSkySpan = D_DrawDoubleSkySpan_16;
		D_DrawSpriteSpan = D_DrawSpriteSpan_16;
		D_DrawFuzzSpriteSpan = D_DrawFuzzSpriteSpan_16;
		D_DrawAltFuzzSpriteSpan = D_DrawFuzzSpriteSpan_16;
	}
	else
	{
		I_Error("Invalid BPP");
	}

    ScaleX = FRACUNIT * ScreenWidth / 320;
    ScaleY = FRACUNIT * ScreenHeight / 200;
    ScaleXI = FRACUNIT * 320 / ScreenWidth;
    ScaleYI = FRACUNIT * 200 / ScreenHeight;
}

//==========================================================================
//
//	InitViewBorder
//
//	Called whenever the view size changes.
//
//==========================================================================

static void InitViewBorder(void)
{ 
    if (r_backscreen)
    {
      	Z_Free(r_backscreen);
		r_backscreen = NULL;
    }

    if (viewwidth == ScreenWidth)
		return;

    r_backscreen = (byte*)Z_Malloc(ScreenPitch * (ScreenHeight - SB_REALHEIGHT) * PixelBytes);

	R_DrawViewBorder();

	memcpy(r_backscreen, scrn, ScreenPitch * (ScreenHeight - SB_REALHEIGHT) * PixelBytes);
} 

//==========================================================================
//
//	VideoErase
//
// 	Copy a screen buffer.
//
//==========================================================================

static void VideoErase(unsigned ofs, int count)
{
	ofs *= PixelBytes;
	count *= PixelBytes;
    memcpy(scrn + ofs, r_backscreen + ofs, count);
} 

//==========================================================================
//
//	EraseViewBorder
//
// 	Draws the border around the view for different size windows
//
//==========================================================================

static void EraseViewBorder(void)
{ 
    int		top;
    int		side;
    int		ofs;
    int		i; 
 
    if (viewwidth == ScreenWidth)
		return;
  
    top = ((ScreenHeight - SB_REALHEIGHT) - viewheight) / 2;
    side = (ScreenWidth - viewwidth) / 2;
 
    // copy top and one line of left side 
    VideoErase(0, top * ScreenPitch + side);
 
    // copy one line of right side and bottom 
    ofs = (viewheight + top - 1) * ScreenPitch + ScreenWidth - side;
    VideoErase(ofs, top * ScreenPitch + side + (ScreenPitch - ScreenWidth));
 
    // copy sides using wraparound 
    ofs = top * ScreenPitch + ScreenWidth - side;
    side = (side << 1) + (ScreenPitch - ScreenWidth);
    
    for (i=1 ; i<viewheight ; i++) 
    { 
		VideoErase(ofs, side);
		ofs += ScreenPitch;
    }
} 

//==========================================================================
//
//	TSoftwareDrawer::StartUpdate
//
//==========================================================================

void TSoftwareDrawer::StartUpdate(void)
{
}

//==========================================================================
//
//	TSoftwareDrawer::SetupView
//
//==========================================================================

void TSoftwareDrawer::SetupView(int x, int y, int width, int height, float fovx, float fovy)
{
    centerxfrac = (float)(width / 2) - 0.5;
    centeryfrac = (float)(height / 2) - 0.5;

	xprojection = (float)(width / 2) / fovx;
	yprojection = (float)(height / 2) / fovy;

    // Preclaculate all row offsets.
    for (int i = 0; i < height; i++)
	{
		ylookup[i] = x + (height - 1 - i + y) * ScreenPitch;
	}

	// draw the border
	InitViewBorder();
}

//==========================================================================
//
//	TSoftwareDrawer::SetupFrame
//
//==========================================================================

void TSoftwareDrawer::SetupFrame(void)
{
	if (viewwidth != ScreenWidth)
	{
		EraseViewBorder();
	}

	// make FDIV fast. This reduces timing precision after we've been running
	// for a while, so we don't do it globally. This also sets chop mode, and
	// we do it here so that setup stuff like the refresh area calculations
	// match what's done in screen.c
	Sys_LowFPPrecision();

	if (cl.fixedcolormap >= 32)
	{
		fixedcolormap = fadetable + cl.fixedcolormap * 256;
		fixedcolormap16 = fadetable16 + cl.fixedcolormap * 256;
	}
	else
	{
		fixedcolormap = 0;
		fixedcolormap16 = 0;
	}
	UpdatePalette();
	D_ClearPolys();
}

//==========================================================================
//
//	TSoftwareDrawer::EndView
//
//==========================================================================

void TSoftwareDrawer::EndView(void)
{
	// back to high floating-point precision
	Sys_HighFPPrecision();
}

//==========================================================================
//
//	TSoftwareDrawer::ReadScreen
//
//==========================================================================

void *TSoftwareDrawer::ReadScreen(int *bpp, bool *bot2top)
{
	void *dst = Z_Malloc(ScreenPitch * ScreenHeight * PixelBytes);
	memcpy(dst, scrn, ScreenPitch * ScreenHeight * PixelBytes);
	*bpp = ScreenBPP;
	*bot2top = false;
	return dst;
}

