//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**	
//**	The 8-bpp span drawing functions.
//**
//**	All drawing to the view buffer is accomplished in this file.
//**  The other refresh files only know about ccordinates, not the
//**  architecture of the frame buffer.
//**
//**	Here find the main potential for optimization, e.g. inline assembly,
//**  different algorithms.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "d_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

#ifndef USEASM

//==========================================================================
//
//	D_DrawSpan8_8
//
//==========================================================================

extern "C" void D_DrawSpan8_8(void)
{
    byte *dest = scrn + ylookup[ds_v] + ds_u;
    int count = ds_count;
	float sdivz = ds_u * d_sdivzstepu + ds_v * d_sdivzstepv + d_sdivzorigin;
	float tdivz = ds_u * d_tdivzstepu + ds_v * d_tdivzstepv + d_tdivzorigin;
	float zi = ds_u * d_zistepu + ds_v * d_zistepv + d_ziorigin;

	float z = (float)FRACUNIT / zi;
	fixed_t s = (int)(sdivz * z) + sadjust;
	fixed_t t = (int)(tdivz * z) + tadjust;

	float sdivzstepu8 = d_sdivzstepu * 8;
	float tdivzstepu8 = d_tdivzstepu * 8;
	float zistepu8 = d_zistepu * 8;

	fixed_t l = ds_l;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			sdivz += sdivzstepu8;
			tdivz += tdivzstepu8;
			zi += zistepu8;

			z = (float)FRACUNIT / zi;
			snext = (int)(sdivz * z) + sadjust;
			tnext = (int)(tdivz * z) + tadjust;

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				sdivz += d_sdivzstepu * spancountminus1;
				tdivz += d_tdivzstepu * spancountminus1;
				zi += d_zistepu * spancountminus1;

				z = (float)FRACUNIT / zi;
				snext = (int)(sdivz * z) + sadjust;
				tnext = (int)(tdivz * z) + tadjust;

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		do
		{
			byte color = ds_source[((t >> 8) & 0xff00) | ((s >> 16) & 0xff)];
			*dest = ds_colormap[color + ((l >> 11) & 0x1f00)];
			dest++;
			l += ds_lstepu;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);

	D_DrawZSpan();
}
#endif

//==========================================================================
//
//	D_DrawMaskedSpan_8
//
//==========================================================================

extern "C" void D_DrawMaskedSpan_8(void)
{
    byte *dest = scrn + ylookup[ds_v] + ds_u;
    word *zdst = zbuffer + ylookup[ds_v] + ds_u;
    int count = ds_count;
	float sdivz = ds_u * d_sdivzstepu + ds_v * d_sdivzstepv + d_sdivzorigin;
	float tdivz = ds_u * d_tdivzstepu + ds_v * d_tdivzstepv + d_tdivzorigin;
	float zi = ds_u * d_zistepu + ds_v * d_zistepv + d_ziorigin;

	float z = (float)FRACUNIT / zi;
	fixed_t s = (int)(sdivz * z) + sadjust;
	fixed_t t = (int)(tdivz * z) + tadjust;

	float sdivzstepu8 = d_sdivzstepu * 8;
	float tdivzstepu8 = d_tdivzstepu * 8;
	float zistepu8 = d_zistepu * 8;

	fixed_t l = ds_l;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	int izi = (int)(zi * 0x8000 * 0x10000);
	int izistepu = (int)(d_zistepu * 0x8000 * 0x10000);

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			sdivz += sdivzstepu8;
			tdivz += tdivzstepu8;
			zi += zistepu8;

			z = (float)FRACUNIT / zi;
			snext = (int)(sdivz * z) + sadjust;
			tnext = (int)(tdivz * z) + tadjust;

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				sdivz += d_sdivzstepu * spancountminus1;
				tdivz += d_tdivzstepu * spancountminus1;
				zi += d_zistepu * spancountminus1;

				z = (float)FRACUNIT / zi;
				snext = (int)(sdivz * z) + sadjust;
				tnext = (int)(tdivz * z) + tadjust;

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		do
		{
			byte color = ds_source[((t >> 8) & 0xff00) | ((s >> 16) & 0xff)];
			if (color)
			{
				*dest = ds_colormap[color + ((l >> 11) & 0x1f00)];
				*zdst = izi >> 16;
			}
			dest++;
			zdst++;
			izi += izistepu;
			l += ds_lstepu;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);
}

//==========================================================================
//
//	D_DrawFuzzSpan_8
//
//==========================================================================

extern "C" void D_DrawFuzzSpan_8(void)
{
    byte *dest = scrn + ylookup[ds_v] + ds_u;
    word *zdst = zbuffer + ylookup[ds_v] + ds_u;
    int count = ds_count;
	float sdivz = ds_u * d_sdivzstepu + ds_v * d_sdivzstepv + d_sdivzorigin;
	float tdivz = ds_u * d_tdivzstepu + ds_v * d_tdivzstepv + d_tdivzorigin;
	float zi = ds_u * d_zistepu + ds_v * d_zistepv + d_ziorigin;

	float z = (float)FRACUNIT / zi;
	fixed_t s = (int)(sdivz * z) + sadjust;
	fixed_t t = (int)(tdivz * z) + tadjust;

	float sdivzstepu8 = d_sdivzstepu * 8;
	float tdivzstepu8 = d_tdivzstepu * 8;
	float zistepu8 = d_zistepu * 8;

	fixed_t l = ds_l;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	int izi = (int)(zi * 0x8000 * 0x10000);
	int izistepu = (int)(d_zistepu * 0x8000 * 0x10000);

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			sdivz += sdivzstepu8;
			tdivz += tdivzstepu8;
			zi += zistepu8;

			z = (float)FRACUNIT / zi;
			snext = (int)(sdivz * z) + sadjust;
			tnext = (int)(tdivz * z) + tadjust;

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				sdivz += d_sdivzstepu * spancountminus1;
				tdivz += d_tdivzstepu * spancountminus1;
				zi += d_zistepu * spancountminus1;

				z = (float)FRACUNIT / zi;
				snext = (int)(sdivz * z) + sadjust;
				tnext = (int)(tdivz * z) + tadjust;

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		do
		{
			word pixelz = izi >> 16;
			if (pixelz >= *zdst)
			{
				byte color = ds_source[((t >> 8) & 0xff00) | ((s >> 16) & 0xff)];
				if (color)
				{
					color = ds_colormap[color + ((l >> 11) & 0x1f00)];
					*dest = ds_transluc[*dest + (color << 8)];
				}
			}
			dest++;
			zdst++;
			izi += izistepu;
			l += ds_lstepu;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);
}

//==========================================================================
//
//	D_DrawAltFuzzSpriteSpan_8
//
//==========================================================================

extern "C" void D_DrawAltFuzzSpan_8(void)
{
    byte *dest = scrn + ylookup[ds_v] + ds_u;
    word *zdst = zbuffer + ylookup[ds_v] + ds_u;
    int count = ds_count;
	float sdivz = ds_u * d_sdivzstepu + ds_v * d_sdivzstepv + d_sdivzorigin;
	float tdivz = ds_u * d_tdivzstepu + ds_v * d_tdivzstepv + d_tdivzorigin;
	float zi = ds_u * d_zistepu + ds_v * d_zistepv + d_ziorigin;

	float z = (float)FRACUNIT / zi;
	fixed_t s = (int)(sdivz * z) + sadjust;
	fixed_t t = (int)(tdivz * z) + tadjust;

	float sdivzstepu8 = d_sdivzstepu * 8;
	float tdivzstepu8 = d_tdivzstepu * 8;
	float zistepu8 = d_zistepu * 8;

	fixed_t l = ds_l;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	int izi = (int)(zi * 0x8000 * 0x10000);
	int izistepu = (int)(d_zistepu * 0x8000 * 0x10000);

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			sdivz += sdivzstepu8;
			tdivz += tdivzstepu8;
			zi += zistepu8;

			z = (float)FRACUNIT / zi;
			snext = (int)(sdivz * z) + sadjust;
			tnext = (int)(tdivz * z) + tadjust;

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				sdivz += d_sdivzstepu * spancountminus1;
				tdivz += d_tdivzstepu * spancountminus1;
				zi += d_zistepu * spancountminus1;

				z = (float)FRACUNIT / zi;
				snext = (int)(sdivz * z) + sadjust;
				tnext = (int)(tdivz * z) + tadjust;

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		do
		{
			word pixelz = izi >> 16;
			if (pixelz >= *zdst)
			{
				byte color = ds_source[((t >> 8) & 0xff00) | ((s >> 16) & 0xff)];
				if (color)
				{
					color = ds_colormap[color + ((l >> 11) & 0x1f00)];
					*dest = ds_transluc[(*dest << 8) + color];
				}
			}
			dest++;
			zdst++;
			izi += izistepu;
			l += ds_lstepu;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);
}

#ifndef USEASM
//==========================================================================
//
//	D_DrawSkySpan_8
//
//==========================================================================

extern "C" void D_DrawSkySpan_8(void)
{
    byte *dest = scrn + ylookup[ds_v] + ds_u;
    int count = ds_count;
	float sdivz = ds_u * d_sdivzstepu + ds_v * d_sdivzstepv + d_sdivzorigin;
	float tdivz = ds_u * d_tdivzstepu + ds_v * d_tdivzstepv + d_tdivzorigin;
	float zi = ds_u * d_zistepu + ds_v * d_zistepv + d_ziorigin;

	float z = (float)FRACUNIT / zi;
	fixed_t s = (int)(sdivz * z) + sadjust;
	fixed_t t = (int)(tdivz * z) + tadjust;

	float sdivzstepu8 = d_sdivzstepu * 8;
	float tdivzstepu8 = d_tdivzstepu * 8;
	float zistepu8 = d_zistepu * 8;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			sdivz += sdivzstepu8;
			tdivz += tdivzstepu8;
			zi += zistepu8;

			z = (float)FRACUNIT / zi;
			snext = (int)(sdivz * z) + sadjust;
			tnext = (int)(tdivz * z) + tadjust;

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				sdivz += d_sdivzstepu * spancountminus1;
				tdivz += d_tdivzstepu * spancountminus1;
				zi += d_zistepu * spancountminus1;

				z = (float)FRACUNIT / zi;
				snext = (int)(sdivz * z) + sadjust;
				tnext = (int)(tdivz * z) + tadjust;

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		do
		{
			byte color = ds_source[((t >> 6) & 0x3fc00) | ((s >> 16) & 0x3ff)];
			*dest = color;
			dest++;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);
}
#endif

//==========================================================================
//
//	D_DrawDoubleSkySpan_8
//
//==========================================================================

extern "C" void D_DrawDoubleSkySpan_8(void)
{
    byte *dest = scrn + ylookup[ds_v] + ds_u;
    int count = ds_count;
	float sdivz = ds_u * d_sdivzstepu + ds_v * d_sdivzstepv + d_sdivzorigin;
	float tdivz = ds_u * d_tdivzstepu + ds_v * d_tdivzstepv + d_tdivzorigin;
	float zi = ds_u * d_zistepu + ds_v * d_zistepv + d_ziorigin;

	float z = (float)FRACUNIT / zi;
	fixed_t s = (int)(sdivz * z) + sadjust;
	fixed_t t = (int)(tdivz * z) + tadjust;

	float sdivzstepu8 = d_sdivzstepu * 8;
	float tdivzstepu8 = d_tdivzstepu * 8;
	float zistepu8 = d_zistepu * 8;

	int spancount;
	fixed_t snext = 0;
	fixed_t tnext = 0;

	fixed_t sstep = 0;
	fixed_t tstep = 0;

	do
   	{
		// calculate s and t at the far end of the span
		if (count >= 8)
		{
			spancount = 8;
		}
		else
		{
			spancount = count;
		}
		count -= spancount;

		if (count)
		{
			sdivz += sdivzstepu8;
			tdivz += tdivzstepu8;
			zi += zistepu8;

			z = (float)FRACUNIT / zi;
			snext = (int)(sdivz * z) + sadjust;
			tnext = (int)(tdivz * z) + tadjust;

			sstep = (snext - s) >> 3;
			tstep = (tnext - t) >> 3;
		}
		else
		{
			if (spancount > 1)
			{
				float spancountminus1 = (float)(spancount - 1);

				sdivz += d_sdivzstepu * spancountminus1;
				tdivz += d_tdivzstepu * spancountminus1;
				zi += d_zistepu * spancountminus1;

				z = (float)FRACUNIT / zi;
				snext = (int)(sdivz * z) + sadjust;
				tnext = (int)(tdivz * z) + tadjust;

				sstep = (snext - s) / (spancount - 1);
				tstep = (tnext - t) / (spancount - 1);
			}
		}

		do
		{
			byte color = ds_source[((t >> 6) & 0x3fc00) | ((s >> 16) & 0x3ff)];
			if (color)
			{
				*dest = color;
			}
			dest++;
			s += sstep;
			t += tstep;
		} while (--spancount);

		s = snext;
		t = tnext;
	}
	while (count > 0);
}

