//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "gl_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//  TOpenGLDrawer::InitLightTables
//
//==========================================================================

void TOpenGLDrawer::InitLightTables(void)
{
    int         i;
    int         j;
	float		light_level;
	float		startmap;

    //	Calculate the light levels to use for each
    // light_level / distance combination.
    for (i = 0; i < LIGHTLEVELS; i++)
    {
        startmap = ((LIGHTLEVELS - 1 - i) * 2) * 256 / LIGHTLEVELS;
        for (j = 0; j < MAXLIGHTZ; j++)
        {
            light_level = startmap - 640.0 / (float)(j + 1);

            if (light_level < 0)
                light_level = 0;

            if (light_level > 255)
                light_level = 255;

            zlight[i][j] = 255 - light_level;
        }
    }
}

//==========================================================================
//
//	TOpenGLDrawer::SetLight
//
//==========================================================================

void TOpenGLDrawer::SetLight(draw_vertex_t &v, float alpha)
{
	float light;
	if (fixedlight)
	{
		light = fixedlight;
	}
	else if (use_zlight)
	{
#if 1
		int lightlevel = (int)(v.l * LIGHTSEGSCALE + extralight);
		if (lightlevel >= LIGHTLEVELS)
			lightlevel = LIGHTLEVELS - 1;
		if (lightlevel < 0)
			lightlevel = 0;

		int index = (int)(Length(v - vieworg) * light_z_scale);
		if (index >= MAXLIGHTZ)
			index = MAXLIGHTZ - 1;

		light = zlight[lightlevel][index];
#else
		float lightlevel = v.l + extralight * 16.0;
		float index = Length(v - vieworg) * light_z_scale;
		light = lightlevel * 2 - 255 + 640.0 / (index + 1);
		if (light < 0)
			light = 0;
		if (light > 255)
			light = 255;
#endif
	}
	else
	{
		light = v.l + extralight * 16.0;
		if (light > 255)
			light = 255;
	}

	light /= 255.0;
	glColor4f(light, light, light, alpha);
}

//==========================================================================
//
//	TOpenGLDrawer::DrawPolygon
//
//==========================================================================

void TOpenGLDrawer::DrawPolygon(draw_vertex_t *cv, int count, int texture,
	int, bool use_fog)
{
	int		i;

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}
	if (use_fog)
	{
		glEnable(GL_FOG);
	}

	glBegin(GL_TRIANGLE_FAN);
	for (i = 0; i < count; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		SetLight(cv[i], 1.0);
		glVertex3f(cv[i].x, cv[i].y, cv[i].z);
	}
	glEnd();

	if (use_fog)
	{
		glDisable(GL_FOG);
	}
}

//==========================================================================
//
//	TOpenGLDrawer::DrawMaskedPolygon
//
//==========================================================================

void TOpenGLDrawer::DrawMaskedPolygon(draw_vertex_t *cv, int count,
	int texture, int translucency, int, bool use_fog)
{
	int		i;

	if (texture & TEXF_FLAT)
	{
		SetFlat(texture);
	}
	else
	{
		SetTexture(texture);
	}
	glEnable(GL_ALPHA_TEST);
	if (use_fog)
	{
		glEnable(GL_FOG);
	}
	if (translucency)
	{
		glEnable(GL_BLEND);
	}

	float alpha = (100.0 - translucency) / 100.0;

	glBegin(GL_TRIANGLE_FAN);
	for (i = 0; i < count; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		SetLight(cv[i], alpha);
		glVertex3f(cv[i].x, cv[i].y, cv[i].z);
	}
	glEnd();

	if (translucency)
	{
		glDisable(GL_BLEND);
	}
	if (use_fog)
	{
		glDisable(GL_FOG);
	}
	glDisable(GL_ALPHA_TEST);
}

//==========================================================================
//
//	TOpenGLDrawer::DrawZPolygon
//
//==========================================================================

void TOpenGLDrawer::DrawZPolygon(draw_vertex_t *cv, int count, int)
{
	int		i;

	glDisable(GL_TEXTURE_2D);

	glBegin(GL_TRIANGLE_FAN);
	for (i = 0; i < count; i++)
	{
		glVertex3f(cv[i].x, cv[i].y, cv[i].z);
	}
	glEnd();

	glEnable(GL_TEXTURE_2D);
}

//==========================================================================
//
//	TOpenGLDrawer::DrawSkyPolygon
//
//==========================================================================

void TOpenGLDrawer::DrawSkyPolygon(draw_vertex_t *cv, int count,
	int texture, bool double_sky)
{
	int		i;

	SetSkyTexture(texture, double_sky);

	//	Sky polys are alredy translated
	glPushMatrix();
	glTranslatef(vieworg.x, vieworg.y, vieworg.z);
	glDisable(GL_DEPTH_TEST);
	if (double_sky)
	{
		glEnable(GL_ALPHA_TEST);
	}

	glBegin(GL_TRIANGLE_FAN);
	glColor4f(1, 1, 1, 1);
	for (i = 0; i < count; i++)
	{
		TVec texpt = cv[i] - r_texorg;
		glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
			DotProduct(texpt, r_taxis) * tex_ih);
		glVertex3f(cv[i].x, cv[i].y, cv[i].z);
	}
	glEnd();

	if (double_sky)
	{
		glDisable(GL_ALPHA_TEST);
	}
	glEnable(GL_DEPTH_TEST);
	glPopMatrix();
}

//==========================================================================
//
//	TOpenGLDrawer::DrawSpritePolygon
//
//==========================================================================

void TOpenGLDrawer::DrawSpritePolygon(draw_vertex_t *cv, int lump,
	int translucency, int translation, bool use_fog)
{
	TVec	texpt;

	SetSpriteLump(lump, translation);

	if (translucency)
	{
		glEnable(GL_BLEND);
	}
	glEnable(GL_ALPHA_TEST);
	if (use_fog)
	{
		glEnable(GL_FOG);
	}

	float alpha = (100.0 - translucency) / 100.0;

	glBegin(GL_QUADS);

	SetLight(cv[0], alpha);
	texpt = cv[0] - r_texorg;
	glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
		DotProduct(texpt, r_taxis) * tex_ih);
	glVertex3f(cv[0].x, cv[0].y, cv[0].z);

	SetLight(cv[1], alpha);
	texpt = cv[1] - r_texorg;
	glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
		DotProduct(texpt, r_taxis) * tex_ih);
	glVertex3f(cv[1].x, cv[1].y, cv[1].z);

	SetLight(cv[2], alpha);
	texpt = cv[2] - r_texorg;
	glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
		DotProduct(texpt, r_taxis) * tex_ih);
	glVertex3f(cv[2].x, cv[2].y, cv[2].z);

	SetLight(cv[3], alpha);
	texpt = cv[3] - r_texorg;
	glTexCoord2f(DotProduct(texpt, r_saxis) * tex_iw,
		DotProduct(texpt, r_taxis) * tex_ih);
	glVertex3f(cv[3].x, cv[3].y, cv[3].z);

	glEnd();

	if (translucency)
	{
		glDisable(GL_BLEND);
	}
	glDisable(GL_ALPHA_TEST);
	if (use_fog)
	{
		glDisable(GL_FOG);
	}
}

